/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2013  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Instance.h"
#include "Libaries.h"
#include "windows.h"

using namespace Scratch;

namespace BromScript {
	extern int _icount = 0;

	int Instance::GetCount(){
		return _icount;
	}

	Instance::Instance(){
		_icount++;

		this->CurrentIncludePath = "";
		this->CurrentStackIndex = 0;
		this->Globals = new Table(this);
		this->KillScriptThreaded = false;
	}

	Instance::~Instance(){
		_icount--;

		while(this->ModuleExits.Count > 0)
			this->ModuleExits.RemoveAt(0)(this);

		while(this->ChunkScopes.Count > 0)
			delete this->ChunkScopes.RemoveAt(0);

		delete this->Globals;

		this->GC.Run();

		while(this->RegisteredUserdataTypes.Count > 0)
			delete this->RegisteredUserdataTypes.RemoveAt(0);
	}

	void Instance::SetVar(CString key, Variable* var){
		if (var == null || var->Type == Variable::VariableType::Null){
			this->Globals->Remove(key);
			return;
		}

		this->Globals->Set(key, var);
	}

	void Instance::RemoveVar(CString key){
		this->Globals->Remove(key);
	}

	Variable* Instance::GetVar(CString key){
		return this->Globals->Get(key);
	}

	Variable* Instance::DoString(const char* codename, const char* code){
		int i = 0;
		while(true){
			if (code[i] == null){
				return this->DoString(codename, code, i);
			}

			i++;
		}
	}

	Variable* Instance::DoString(const char* codename, const char* code, int size){
		Compiler c(codename, code, size);
		byte* bytecode = c.Run(&size);

		Variable* ret = this->DoCode(codename, bytecode, size);
		delete bytecode;

		return ret;
	}

	void Instance::EnterStack(Function* func){
		CallStack* cs = new CallStack();
		cs->Func = func;
		cs->Name = func->Name;
		cs->Filename = func->Filename;

		this->CurrentStack.Add(cs);
		this->CurrentStackIndex++;
	}

	void Instance::LeaveStack(){
		this->CurrentStackIndex--;
		delete this->CurrentStack.RemoveAt(this->CurrentStackIndex);

		if (this->CurrentStackIndex == 0)
			this->KillScriptThreaded = false;
	}

	int Instance::GetCurrentStackSize(){
		return this->CurrentStackIndex;
	}

	void Instance::DoDLL(const char* filename){
		CString file(filename);
		if (file.Size() < 2 || file[1] != ':')
			file = this->CurrentIncludePath + filename;

		HINSTANCE dll = LoadLibrary(file);
		if (dll == null){
			this->Error(CString("Could not find module '") + file + "'");
			return;
		}
		
		BSModuleFunction init = (BSModuleFunction)GetProcAddress(dll, "BromScriptStart");
		BSModuleFunction deinit = (BSModuleFunction)GetProcAddress(dll, "BromScriptShutdown");
		if (init == null){
			this->Error(CString("Could not find entry point ('BromScriptStart') in module '") + file + "'");
			return;
		}

		if (deinit == null){
			this->Error(CString("Could not find exit point ('BromScriptShutdown') in module '") + file + "'");
			return;
		}

		init(this);
		this->ModuleExits.Add(deinit);
	}

	Variable* Instance::DoFile(const char* filename){
		CString oldincp = this->CurrentIncludePath;
		CString file(filename);
		file = file.Replace('\\', '/');

		if (file.Size() > 2 && file[1] != ':')
			file = this->CurrentIncludePath + filename;

		if (file.Contains('/'))
			this->CurrentIncludePath += file.Substring(0, file.LastIndexOf('/') + 1);

		CFileStream fs;
		fs.Open(file, "rb");

		int filesize = fs.Size();
		byte* buff = (byte*)fs.Read(filesize);
		fs.Close();

		if (filesize > 13){
			// \0 BROMSCRIPT \0 VERSION
			bool iscompiled = true;
			byte bytecheck[] = {0, 66, 82, 79, 77, 83, 67, 82, 73, 80, 84, 0};
			for (int i = 0; i < 12; i++){ // tag is 13, but the last one is version so go to 12, we'll handle the version later.
				if (bytecheck[i] != buff[i]){
					iscompiled = false;
					break;
				}
			}

			if (iscompiled){
				if (buff[12] != BROMSCRIPT_CURRENTVERSION){
					this->Error(CString("Cannot execute '") + filename + "', invalid CBS version.");
					delete buff;
					return null;
				}

				for (int i = 0; i < 13; i++)
					buff[i] = (byte)Misc::ExecFuncs::Skip;

				Variable* ret = this->DoCode(filename, buff, filesize);
				delete buff;

				this->CurrentIncludePath = oldincp;
				return ret;
			}
		}

		Compiler c(filename, (char*)buff, filesize);
		byte* bytecode = c.Run(&filesize);
		delete buff;

		Variable* ret = this->DoCode(filename, bytecode, filesize);
		delete bytecode;

		this->CurrentIncludePath = oldincp;
		return ret;
	}

	Variable* Instance::DoCode(const char* filename, byte* code, int codelen){
		Function* f = new Function(this);
		f->Code = new byte[codelen];
		f->Name = "C++_DoCode";
		f->Filename = filename;

		memcpy(f->Code, code, codelen);

		Variable* ret = f->Run();

		this->ChunkScopes.Add(f);
		return ret;
	}

	Function* Instance::GetCurrentFunction(){
		if (this->CurrentStackIndex == 0) return null;
		return this->CurrentStack[this->CurrentStackIndex - 1]->Func;
	}

	CallStack* Instance::GetCurrentStack(){
		CallStack* stack = new CallStack[this->CurrentStackIndex];
		for (int i = 0; i < this->CurrentStackIndex; i++)
			stack[i] = *this->CurrentStack[i];

		return stack;
	}

	void Instance::Error(CString msg){
		if (this->ErrorCallback == null)
			throw new _exception();

		CallStack* stack = new CallStack[this->CurrentStackIndex];
		for (int i = 0; i < this->CurrentStackIndex; i++){
			this->CurrentStack[i]->Func->ForceReturn = true;
			this->CurrentStack[i]->LineNumber = this->CurrentStack[i]->Func->CurrentSourceFileLine;
			stack[i] = *this->CurrentStack[i];
		}

		this->ErrorCallback(this, stack, this->CurrentStackIndex, msg);

		delete[] stack;
	}

	void Instance::RegisterScope(Function* func){
		this->ChunkScopes.Add(func);
	}

	void Instance::SetErrorCallback(ErrorFunction func){
		this->ErrorCallback = func;
	}

	void Instance::RegisterFunction(CString key, BSFunction func){
		Function* varfunc = new Function(this);
		varfunc->CppFunc = func;
		varfunc->Filename = "C++";
		varfunc->IsCpp = true;
		varfunc->Name = key;

		Variable* var = this->GC.RegisterVariable();
		var->Type = Variable::VariableType::Function;
		var->Value = varfunc;

		this->Globals->Set(key, var);
	}

	Variable* Instance::CreateUserdata(int typeID, void* ptr, bool calldtor){
		Userdata* ud = this->GetRegisteredUserdata(typeID)->Copy();
		ud->TypeData = ptr;
		ud->CallDTor = calldtor;
		
		Variable* var = new Variable();
		var->Value = ud;
		var->Type = (Variable::VariableType)typeID;

		return var;
	}

	Userdata* Instance::GetRegisteredUserdata(int typeID){
		for (int i = 0; i < this->RegisteredUserdataTypes.Count; i++){
			if (this->RegisteredUserdataTypes[i]->TypeID == typeID){
				return this->RegisteredUserdataTypes[i];
			}
		}	

		return null;
	}

	Userdata* Instance::RegisterUserdata(CString name, int typeID, int typesize, BSFunctionCtor ctor, BSFunctionDtor dtor){
		if (typeID < 100)
			throw "Cannot register Userdata with type below 100! (reserved for internal types)";

		for(int i = 0; i < this->RegisteredUserdataTypes.Count; i++){
			if (this->RegisteredUserdataTypes[i]->TypeID == typeID){
				CString str = CString("Userdata typeID ") + CString((double)typeID) + CString(" already registered under the name ") + this->RegisteredUserdataTypes[i]->Name;
				throw str.str_szBuffer;
			}
		}

		Userdata* ud = new Userdata();
		ud->Name = name;
		ud->Ctor = ctor;
		ud->Dtor = dtor;
		ud->TypeID = typeID;
		ud->BromScript = this;
		ud->TypeSize = typesize;

		this->RegisteredUserdataTypes.Add(ud);

		Function* func = new Function(this);
		func->Name = name;
		func->IsCpp = true;

		Variable* classvar = this->GC.RegisterVariable();
		classvar->Type = Variable::VariableType::Class;
		classvar->Value = func;

		this->SetVar(name, classvar);

		return ud;
	}

	void Instance::LoadDefaultLibaries(){
		this->RegisterFunction("setReadOnly", BromScript::Libaries::Global::SetReadOnly);
		this->RegisterFunction("tostring", BromScript::Libaries::Global::ToString);
		this->RegisterFunction("print", BromScript::Libaries::Global::Print);
		this->RegisterFunction("call", BromScript::Libaries::Global::Call);
		this->RegisterFunction("typeof", BromScript::Libaries::Global::TypeOf);
		this->RegisterFunction("include", BromScript::Libaries::Global::Include);
		this->RegisterFunction("includeDll", BromScript::Libaries::Global::IncludeDll);
		this->RegisterFunction("error", BromScript::Libaries::Global::Error);

		Table* gc = new Table(this);
		gc->Set("Run", Converter::ToVariable(this, "Run", BromScript::GarbageCollector::RunWrapper));
		this->SetVar("GC", Converter::ToVariable(gc));

		Table* math = new Table(this);
		math->Set("Random", Converter::ToVariable(this, "Random", BromScript::Libaries::Math::Random));
		math->Set("Floor", Converter::ToVariable(this, "Floor", BromScript::Libaries::Math::Floor));
		math->Set("Min", Converter::ToVariable(this, "Min", BromScript::Libaries::Math::Min));
		math->Set("Max", Converter::ToVariable(this, "Max", BromScript::Libaries::Math::Max));
		math->Set("Ceiling", Converter::ToVariable(this, "Ceiling", BromScript::Libaries::Math::Ceiling));
		math->Set("Round", Converter::ToVariable(this, "Round", BromScript::Libaries::Math::Round));
		math->Set("Atan", Converter::ToVariable(this, "Atan", BromScript::Libaries::Math::Atan));
		math->Set("Atan2", Converter::ToVariable(this, "Atan2", BromScript::Libaries::Math::Atan2));
		math->Set("Cos", Converter::ToVariable(this, "Cos", BromScript::Libaries::Math::Cos));
		math->Set("Sin", Converter::ToVariable(this, "Sin", BromScript::Libaries::Math::Sin));
		math->Set("Tan", Converter::ToVariable(this, "Tan", BromScript::Libaries::Math::Tan));
		math->Set("Pow", Converter::ToVariable(this, "Pow", BromScript::Libaries::Math::Pow));
		math->Set("Sign", Converter::ToVariable(this, "Sign", BromScript::Libaries::Math::Sign));
		math->Set("Sqrt", Converter::ToVariable(this, "Sqrt", BromScript::Libaries::Math::Sqrt));
		math->Set("Abs", Converter::ToVariable(this, "Abs", BromScript::Libaries::Math::Abs));
		math->Set("Neg", Converter::ToVariable(this, "Neg", BromScript::Libaries::Math::Neg));
		math->Set("ToRadians", Converter::ToVariable(this, "ToRadians", BromScript::Libaries::Math::ToRadians));
		math->Set("ToDegrees", Converter::ToVariable(this, "ToDegrees", BromScript::Libaries::Math::ToDegrees));
		math->Set("Pi", Converter::ToVariable(BS_PI));
		this->SetVar("Math", Converter::ToVariable(math));

		Table* console = new Table(this);
		console->Set("SetForeColor", Converter::ToVariable(this, "SetForeColor", BromScript::Libaries::Console::SetForeColor));
		console->Set("SetBackColor", Converter::ToVariable(this, "SetBackColor", BromScript::Libaries::Console::SetBackColor));
		console->Set("SetDefaultColors", Converter::ToVariable(this, "SetDefaultColors", BromScript::Libaries::Console::SetDefaultColors));
		console->Set("ColorBlack", Converter::ToVariable((double)BromScript::Libaries::Console::ColorBlack));
		console->Set("ColorBlue", Converter::ToVariable((double)BromScript::Libaries::Console::ColorBlue));
		console->Set("ColorCyan", Converter::ToVariable((double)BromScript::Libaries::Console::ColorCyan));
		console->Set("ColorDarkBlue", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkBlue));
		console->Set("ColorDarkCyan", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkCyan));
		console->Set("ColorDarkGreen", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkGreen));
		console->Set("ColorDarkMagenta", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkMagenta));
		console->Set("ColorDarkRed", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkRed));
		console->Set("ColorDarkWhite", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkWhite));
		console->Set("ColorDarkYellow", Converter::ToVariable((double)BromScript::Libaries::Console::ColorDarkYellow));
		console->Set("ColorGray", Converter::ToVariable((double)BromScript::Libaries::Console::ColorGray));
		console->Set("ColorGreen", Converter::ToVariable((double)BromScript::Libaries::Console::ColorGreen));
		console->Set("ColorMagenta", Converter::ToVariable((double)BromScript::Libaries::Console::ColorMagenta));
		console->Set("ColorRed", Converter::ToVariable((double)BromScript::Libaries::Console::ColorRed));
		console->Set("ColorWhite", Converter::ToVariable((double)BromScript::Libaries::Console::ColorWhite));
		console->Set("ColorYellow", Converter::ToVariable((double)BromScript::Libaries::Console::ColorYellow));
		this->SetVar("Console", Converter::ToVariable(console));

		Table* string = new Table(this);
		string->Set("ToChar", Converter::ToVariable(this, "ToChar", BromScript::Libaries::String::ToChar));
		string->Set("FromChar", Converter::ToVariable(this, "FromChar", BromScript::Libaries::String::FromChar));
		string->Set("Split", Converter::ToVariable(this, "Split", BromScript::Libaries::String::Split));
		string->Set("Sub", Converter::ToVariable(this, "Sub", BromScript::Libaries::String::Sub));
		string->Set("IndexOf", Converter::ToVariable(this, "IndexOf", BromScript::Libaries::String::IndexOf));
		string->Set("LastIndexOf", Converter::ToVariable(this, "LastIndexOf", BromScript::Libaries::String::LastIndexOf));
		string->Set("Left", Converter::ToVariable(this, "Left", BromScript::Libaries::String::Left));
		string->Set("Right", Converter::ToVariable(this, "Right", BromScript::Libaries::String::Right));
		string->Set("Trim", Converter::ToVariable(this, "Trim", BromScript::Libaries::String::Trim));
		string->Set("TrimLeft", Converter::ToVariable(this, "TrimLeft", BromScript::Libaries::String::TrimLeft));
		string->Set("TrimRight", Converter::ToVariable(this, "TrimRight", BromScript::Libaries::String::TrimRight));
		string->Set("Replace", Converter::ToVariable(this, "Replace", BromScript::Libaries::String::Replace));
		this->SetVar("String", Converter::ToVariable(string));
	}
}